"""Generate and edit metadata based on an input Spark DataFrame."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
import pyspark.sql.functions as F
import pyspark.sql.types as T
from dateutil.parser import ParserError
from ipydatagrid import DataGrid  # type: ignore[import, import-untyped]
from IPython.display import display
from ipywidgets import Button, HBox, VBox  # type: ignore[import, import-untyped]
from loguru import logger
from pandas.core.tools.datetimes import (  # type: ignore[attr-defined]
    _guess_datetime_format_for_array,
)
from pyspark.sql import DataFrame, Row, SparkSession
from pyspark.sql.window import Window

from pushcart.metadata.spark import generate_code as sp_generate_code
from pushcart.metadata.spark import transform as sp_transform
from pushcart.utils import pandas_to_spark_datetime_pattern

if TYPE_CHECKING:
    from collections.abc import Hashable


def _infer_json_schema(
    df: DataFrame,
    column_name: str,
    spark: SparkSession | None = None,
) -> str | None:
    logger.info(f"Attempting to infer JSON schema for {column_name} column.")

    if not spark:
        spark = SparkSession.builder.getOrCreate()
    schema = spark.read.json(df.select(column_name).rdd.map(lambda x: x[0])).schema

    if schema.fieldNames() == ["_corrupt_record"]:
        logger.warning(f"Could not infer JSON schema for {column_name} column.")
        return None

    if "_corrupt_record" in schema.fieldNames():
        schema.fields.pop(
            schema.fields.index(
                T.StructField("_corrupt_record", T.StringType(), nullable=True),
            ),
        )

    return f'F.from_json(F.col("{column_name}"), schema="{schema.simpleString()}")'


def _infer_timestamps(df: DataFrame, column_name: str) -> str | None:
    logger.info(f"Attempting to infer timestamp format for {column_name} column.")

    pd_ts_format = None

    with contextlib.suppress(ParserError, AttributeError, TypeError):
        pd_ts_df = df.select(column_name).dropna().limit(1).toPandas()
        # Pandas keeps the last bit of the field path as column name
        pd_ts_list = pd_ts_df[column_name.split(".")[-1]].to_numpy()

        pd_ts_format = _guess_datetime_format_for_array(pd_ts_list)

    if not pd_ts_format:
        logger.warning(f"Could not infer timestamp format for {column_name} column.")
        return None

    spark_ts_format = pandas_to_spark_datetime_pattern(pd_ts_format)

    return f'F.to_timestamp(F.col("{column_name}"), "{spark_ts_format}")'


def get_unique_values(
    df: DataFrame,
    max_rows: int | None = None,
    spark: SparkSession | None = None,
) -> DataFrame:
    """Get a dataframe with only per-column unique values for inference.

    Useful for more accurately inferring metadata for some datasets.
    Can take a long time to run, but will also reduce the runtime for
    `Metadata` generation if used as its input.

    This method will break the inter-column consistency of records, since
    the distinct() operation is performed per each column. For example:

    ```
    [
        { "a": 1, "b": 5 },
        { "a": 1, "b": 6 },
        { "a": 2, "b": 7 },
        { "a": 3, "b": 7 },
        { "a": 4, "b": 7 }
    ]
    ```

    Will yield:

    ```
    [
        { "a": 1, "b": 5 },
        { "a": 2, "b": 6 },
        { "a": 3, "b": 7 },
        { "a": 4, "b": None }
    ]
    ```

    Parameters
    ----------
    df : DataFrame
        Spark DataFrame containing data to be analyzed
    max_rows : int, optional
        Number of unique values to bring in per column, by default None.
        If not specified, max_rows will use the total record count of `df`
    spark: SparkSession, optional
        Optionally provide an existing SparkSession, by default None

    Returns
    -------
    DataFrame
        Spark DataFrame containing per-column unique data

    """
    if not df:
        msg = "Parameter `df` must be a Spark DataFrame"
        raise ValueError(msg)

    if not spark:
        spark = SparkSession.builder.getOrCreate()

    if df.isEmpty():
        return spark.createDataFrame([], df.schema)

    if not max_rows:
        max_rows = df.count() or 0

    unique_vals_df: DataFrame = spark.createDataFrame(
        [Row(_row_index=r) for r in range(1, max_rows)],
    )

    for c in df.columns:
        unique_vals_df = unique_vals_df.join(
            df.select(c)
            .distinct()
            .limit(max_rows)
            .withColumn(
                "_row_index",
                F.row_number()
                .over(
                    Window.orderBy(F.monotonically_increasing_id()),
                )
                .alias(f"{c}_col"),
            ),
            on=["_row_index"],
            how="left_outer",
        )

    return unique_vals_df.select(
        [col for col in unique_vals_df.columns if not col.startswith("_")],
    )


class Metadata:
    """Generate metadata and transformations based on input DataFrame."""

    def __init__(
        self,
        df: DataFrame | None,
        infer_json_schema: bool = True,
        infer_timestamps: bool = True,
        infer_fraction: float = 0.25,
    ) -> None:
        """Generate metadata and transformations based on input DataFrame.

        Parameters
        ----------
        df : DataFrame
            PySpark DataFrame for which metadata is generated
        infer_json_schema : bool, optional
            Whether to infer the schema of JSON string columns or not, by default True
        infer_timestamps : bool, optional
            Whether to try to infer timestamp string column formats, by default True
        infer_fraction : float, optional
            The fraction of data that is sampled from the input, by default 0.25

        """
        self.data_df = df

        self.infer_json_schema = infer_json_schema
        self.infer_timestamps = infer_timestamps
        self.infer_fraction = infer_fraction

        self.metadata_df: pd.DataFrame | None = None
        self.metadata_grid: DataGrid = None

        self.metadata_cols = [
            "column_order",
            "source_column_name",
            "source_column_type",
            "dest_column_name",
            "dest_column_type",
            "transform_function",
            "default_value",
            "validation_rule",
            "validation_action",
        ]

    @classmethod
    def from_existing_csv(cls, path: str, data_df: DataFrame | None = None) -> Metadata:
        """Load metadata from an existing metadata file, without the original dataset.

        Parameters
        ----------
        path : str
            Path to metadata file.
        data_df : DataFrame | None, optional
            Optionally, attach an actual dataset to the Metadata object for running
            transformations on it automatically, by default None

        Returns
        -------
        Metadata
            Object allowing to visualize, edit, and generate PySpark code from metadata.

        """
        md = Metadata(df=None)
        md.metadata_df = pd.read_csv(path)
        md.data_df = data_df

        return md

    @classmethod
    def from_df(
        cls,
        metadata_df: pd.DataFrame | DataFrame,
        data_df: DataFrame | None = None,
    ) -> Metadata:
        """Load metadata from an in-memory DataFrame, without the original dataset.

        Parameters
        ----------
        metadata_df : pd.DataFrame | DataFrame
            DataFrame object containing metadata and transformations.
        data_df : DataFrame | None, optional
            Optionally, attach an actual dataset to the Metadata object for running
            transformations on it automatically, by default None

        Returns
        -------
        Metadata
            Object allowing to visualize, edit, and generate PySpark code from metadata.

        Raises
        ------
        TypeError
            metadata_df must be a pandas or Spark DataFrame

        """
        md = Metadata(df=None)
        if not isinstance(metadata_df, pd.DataFrame | DataFrame):
            msg = "metadata_df must be a pandas or Spark DataFrame"
            raise TypeError(msg)
        md.metadata_df = (
            pd.DataFrame(metadata_df.toPandas()).fillna(value=np.nan)
            if isinstance(metadata_df, DataFrame)
            else metadata_df
        )
        md.data_df = data_df

        return md

    def _infer(self, column_name: str) -> str:
        if not self.data_df:
            msg = "Cannot infer metadata since no Spark DataFrame has been provided."
            raise ValueError(msg)

        sampled_df = self.data_df.sample(
            fraction=self.infer_fraction,
            withReplacement=False,
        )

        inferred_ts = None
        inferred_schema = None

        if self.infer_timestamps:
            inferred_ts = _infer_timestamps(sampled_df, column_name)
            if inferred_ts:
                return inferred_ts

        if self.infer_json_schema:
            inferred_schema = _infer_json_schema(sampled_df, column_name)
            if inferred_schema:
                return inferred_schema

        return ""

    def _generate_field(self, field: T.StructField, parent: str | None = None) -> dict:
        name = field.name
        dtype = field.dataType
        flat_parent = parent.replace(".", "_") if parent else None
        return {
            "source_column_name": f"{flat_parent}.{name}" if parent else name,
            "source_column_type": dtype.simpleString(),
            "dest_column_name": f"{flat_parent}_{name}" if parent else name,
            "dest_column_type": (
                dtype.elementType.simpleString()
                if isinstance(dtype, T.ArrayType)
                else dtype.simpleString()
            ),
            "transform_function": (
                (
                    f'F.explode("{flat_parent}.{name}")'
                    if parent
                    else f'F.explode("{name}")'
                )
                if isinstance(dtype, T.ArrayType)
                else (
                    self._infer(f"{parent}.{name}" if parent else name)
                    if isinstance(dtype, T.StringType)
                    else ""
                )
            ),
            "default_value": "",
            "validation_rule": "",
            "validation_action": "",
        }

    def _generate(self, schema: T.StructType, parent: str | None = None) -> list:
        fields = []

        for f in schema.fields:
            dtype = f.dataType

            field_name = f"{parent + '.' if parent else ''}{f.name}"

            if isinstance(dtype, T.ArrayType) and isinstance(
                dtype.elementType,
                T.StructType,
            ):
                fields.append(self._generate_field(f, parent))
                fields += self._generate(
                    dtype.elementType,
                    parent=field_name,
                )
            elif isinstance(dtype, T.StructType):
                fields += self._generate(dtype, parent=field_name)
            else:
                fields.append(self._generate_field(f, parent))

        return fields

    def generate(self) -> pd.DataFrame:
        """Generate a Pandas DataFrame containing the metadata for the dataset being analyzed.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame containing proposed transformation metadata

        """
        if self.data_df is None:
            msg = "Cannot generate new metadata since no Spark DataFrame has been provided."
            raise ValueError(msg)

        metadata = pd.DataFrame(
            self._generate(self.data_df.schema, parent=None),
            columns=self.metadata_cols,
        )

        metadata["column_order"] = metadata.index

        return metadata

    def _on_cell_changed(self, _: dict) -> None:
        self.metadata_df = self.metadata_grid.data

    def _on_add_row(self, _: Button) -> None:
        data = self.metadata_grid.get_visible_data()
        new_row = pd.DataFrame(
            columns=data.columns,
            index=[data.index[-1] + 1],
            data="" * len(data.columns),
        )
        self.metadata_grid.data = pd.concat([data, new_row]).reset_index(drop=True)

    def _on_remove_row(self, _: Button) -> None:
        data = self.metadata_grid.get_visible_data()
        self.metadata_grid.data = data.drop(data.tail(1).index)

    def visualize(self) -> None:
        """Show the generated metadata and allow user editing.

        Raises
        ------
        ValueError
            Can only show the metadata if it is present in memory.

        """
        if self.metadata_df is None:
            msg = "Cannot visualize metadata as it has not yet been generated/loaded."
            raise ValueError(msg)

        self.metadata_grid = DataGrid(self.metadata_df, editable=True)
        self.metadata_grid.auto_fit_params = {"area": "all"}
        self.metadata_grid.auto_fit_columns = True
        self.metadata_grid.on_cell_change(self._on_cell_changed)

        add_row_button = Button(description="Add Row")
        add_row_button.on_click(self._on_add_row)

        remove_row_button = Button(description="Remove Row")
        remove_row_button.on_click(self._on_remove_row)

        layout = VBox([HBox([add_row_button, remove_row_button]), self.metadata_grid])

        display(layout)

    def get_metadata(self, show: bool = True) -> None:
        """Generate new metadata for the provided dataset and visualize it.

        Parameters
        ----------
        show : bool
            Display metadata editor in a notebook.

        """
        self.metadata_df = self.generate()

        if show:
            self.visualize()

    def save_to_csv(self, path: str) -> None:
        """Save the current version of the metadata to a CSV file.

        Parameters
        ----------
        path : str
            File path to save to

        """
        if not self.metadata_df:
            msg = "Cannot save metadata as it has not yet been generated/loaded."
            raise ValueError(msg)

        self.metadata_df.to_csv(path)
        logger.info(f"Wrote {len(self.metadata_df.index)} lines to {path}")

    @staticmethod
    def _keep_dest_cols(metadata_df: pd.DataFrame) -> pd.DataFrame:
        has_dest_col = (metadata_df["dest_column_name"].isna()) | (
            metadata_df["dest_column_name"] == ""
        )
        return metadata_df.loc[~has_dest_col]

    @staticmethod
    def _drop_technical_cols(metadata_df: pd.DataFrame) -> pd.DataFrame:
        tech_cols = metadata_df["dest_column_name"].str.startswith("_")

        excluded_columns = (
            metadata_df.loc[tech_cols]["dest_column_name"].to_list() or None
        )
        logger.info(f"Excluding technical columns: {excluded_columns}")

        return metadata_df.loc[~tech_cols]

    def _prepare_metadata(
        self,
        keep_technical_cols: bool = False,
    ) -> tuple[list[dict[Hashable, Any]], list[str]]:
        """Parse the metadata and keep or drop technical columns (starting with underscore).

        Parameters
        ----------
        keep_technical_cols : bool, optional
            Generate transformations for columns whose names start with underscore, by default False

        Returns
        -------
        tuple[pd.DataFrame, list[str]]
            DataFrame with the final set of pending transformations and list of columns to keep.

        Raises
        ------
        ValueError
            Can only generate code based on metadata that is present in memory.

        """
        if self.metadata_df is None:
            msg = "Cannot generate PySpark code from empty metadata."
            raise ValueError(msg)

        mdf = self._keep_dest_cols(self.metadata_df)

        dest_cols = [col for col in mdf["dest_column_name"].to_list() if col] or None
        if not dest_cols:
            logger.warning(
                "No destination columns defined in metadata. Using source column names.",
            )
            dest_cols = [col for col in mdf["source_column_name"].to_list() if col]

        if not dest_cols:
            msg = "Cannot generate PySpark code from empty metadata."
            raise ValueError(msg)

        if not keep_technical_cols:
            mdf = self._drop_technical_cols(mdf)

        return mdf.sort_values(by=["column_order"]).to_dict(orient="records"), dest_cols

    def generate_code(self, keep_technical_cols: bool = False) -> str | None:
        """Generate PySpark transformation code based on existing metadata.

        Parameters
        ----------
        keep_technical_cols : bool, optional
            Generate transformations for columns whose names start with underscore, by default False

        Returns
        -------
        str | None
            String containing runnable PySpark code

        """
        transformations, dest_cols = self._prepare_metadata(keep_technical_cols)

        return sp_generate_code(transformations, dest_cols)

    def transform(self, keep_technical_cols: bool = False) -> DataFrame:
        """Perform the transformations configured in the metadata table.

        Parameters
        ----------
        keep_technical_cols : bool, optional
            Generate transformations for columns whose names start with underscore, by default False

        Returns
        -------
        DataFrame
            Spark DataFrame containing the transformed data

        """
        if not self.data_df:
            msg = "Cannot perform transformations since no Spark DataFrame has been provided."
            raise ValueError(msg)

        transformations, dest_cols = self._prepare_metadata(keep_technical_cols)

        return sp_transform(self.data_df, transformations, dest_cols)
